==> 1.1.4-dev.tar.gz: lib/src/x_http_client.dart <==
import 'dart:async';
import 'dart:convert';
import 'dart:typed_data';

import 'package:flutterx_http/flutterx_http.dart';
import 'package:flutterx_http/src/transform_request.dart';
import 'package:flutterx_utils/flutterx_utils.dart';
import 'package:http/http.dart';
import 'package:http/src/utils.dart'; // ignore: implementation_imports
import 'package:http_parser/http_parser.dart';

part 'x_http_exception.dart';
part 'x_http_extensions.dart';

typedef OnRequest = FutureOr<BaseRequest> Function(BaseRequest request);
typedef OnResponse = FutureOr<StreamedResponse> Function(StreamedResponse response);
typedef OnError = FutureOr<BaseRequest?> Function(XHTTPException error);
typedef OnMultipartFileProgress = void Function(int fileIndex, int length, int total);

/// Type alias for query parameters map
typedef QueryParameters = Map<String, dynamic>;

/// Type alias for headers map
typedef HeadersMap = Map<String, String>;

class XHttpClient with Loggable {
  final Client _inner;

  /// Used for building requests url
  final String? baseUrl;

  final OnRequest? onRequest;
  final OnResponse? onResponse;
  final OnError? onError;

  /// Creates a XHttpClient from [client], defaults to Client()
  XHttpClient({Client? client, this.baseUrl, this.onRequest, this.onResponse, this.onError})
      : _inner = client ?? Client();

  Future<Response> head(String url, {QueryParameters params = const {}, HeadersMap headers = const {}}) =>
      send(applicationJsonRequest('HEAD', buildUrl(url, params: params), headers: headers));

  Future<Response> get(String url, {QueryParameters params = const {}, HeadersMap headers = const {}}) =>
      send(applicationJsonRequest('GET', buildUrl(url, params: params), headers: headers));

  Future<Response> post(String url,
          {QueryParameters params = const {}, HeadersMap headers = const {}, Object? body, Encoding? encoding}) =>
      send(applicationJsonRequest('POST', buildUrl(url, params: params),
          headers: headers, body: body, encoding: encoding));

  Future<Response> put(String url,
          {QueryParameters params = const {}, HeadersMap headers = const {}, Object? body, Encoding? encoding}) =>
      send(applicationJsonRequest('PUT', buildUrl(url, params: params),
          headers: headers, body: body, encoding: encoding));

  Future<Response> patch(String url,
          {QueryParameters params = const {}, HeadersMap headers = const {}, Object? body, Encoding? encoding}) =>
      send(applicationJsonRequest('PATCH', buildUrl(url, params: params),
          headers: headers, body: body, encoding: encoding));

  Future<Response> delete(String url, {QueryParameters params = const {}, HeadersMap headers = const {}}) =>
      send(applicationJsonRequest('DELETE', buildUrl(url, params: params), headers: headers));

  Future<Response> multipart(
    String url, {
    String method = 'POST',
    QueryParameters params = const {},
    HeadersMap headers = const {},
    Map<String, String> fields = const {},
    List<MultipartFile> files = const [],
    OnMultipartFileProgress? onProgress,
  }) {
    BaseRequest request = MultipartRequest(method, buildUrl(url, params: params))
      ..headers.addAll(headers)
      ..headers.putIfAbsent('content-type', () => 'multipart/form-data')
      ..fields.addAll(fields)
      ..files.addAll(files);
    if (onProgress != null) {
      var byteCount = 0;
      var fileIndex = 0;
      request = request.observeProgress((readCount) {
        if (fileIndex >= files.length) return;
        byteCount += readCount;
        final length = files[fileIndex].length;
        if (byteCount > length) {
          onProgress(fileIndex, length, length);
          byteCount -= length;
          fileIndex++;
        } else {
          onProgress(fileIndex, byteCount, length);
        }
      });
    }
    return send(request);
  }

  Future<Response> send(BaseRequest request) async {
    if (onRequest != null) request = await onRequest!(request);
    _printRequest(request);
    var stream = await _inner.send(request);
    if (onResponse != null) stream = await onResponse!(stream);
    final response = await Response.fromStream(stream);
    _printResponse(response);
    if (response.isSuccess) return response;
    final error = XHTTPException(response);
    final retry = onError == null ? null : await onError!(error);
    if (retry != null) return send(retry);
    throw error;
  }

  void close() => _inner.close();

  Uri buildUrl(String url, {QueryParameters params = const {}}) =>
      buildUrlBase(baseUrl: baseUrl, url: url, params: params);

  void _printRequest(BaseRequest request) => log('HTTP Request', () {
        logKV('method', request.method);
        logKV('url', '${request.url.scheme}://${request.url.host}:${request.url.port}${request.url.path}');
        final parameters = request.url.queryParameters;
        if (parameters.isNotEmpty) log('parameters:', () => parameters.forEach(logKV));
        if (request.headers.isNotEmpty) log('headers:', () => request.headers.forEach(logKV));
        if (request.contentLength != null) logKV('contentLength', request.contentLength);
        void _printRequestBody(BaseRequest request) {
          if (request is Request) {
            final body = request.bodyBytes;
            if (body.isNotEmpty) log('body:', () => log(body));
          } else if (request is StreamedRequest) {
            logKV('body', 'stream');
          } else if (request is MultipartRequest) {
            log('body:', () {
              if (request.fields.isNotEmpty) log('fields', () => request.fields.forEach(logKV));
              if (request.files.isNotEmpty) log('files', () => request.files.map((file) => file.filename).forEach(log));
            });
          } else if (request is TransformRequest) {
            logKV('transform', true);
            _printRequestBody(request.source);
          }
        }

        _printRequestBody(request);
      });

  void _printResponse(BaseResponse response) => log('HTTP Response', () {
        final request = response.request;
        if (request != null) {
          logKV('method', request.method);
          logKV('url', request.url);
        }
        logKV('statusCode', response.statusCode);
        logKV('reasonPhrase', response.reasonPhrase);
        if (response.headers.isNotEmpty) log('headers:', () => response.headers.forEach(logKV));
        if (response is Response) {
          final body = response.body;
          if (body.isNotEmpty) log('body:', () => log(body));
        } else if (response is StreamedResponse) {
          log('body: stream');
        }
      });

  /// Creates an HTTP application/json request
  ///
  /// The content-type of the request will default to 'application/json' if not set.
  ///
  /// If provided, [body] sets the body of the request. It can be a [String], any [Object].
  ///
  /// If [body] is a String, it's encoded using [encoding] and used as the body of the request
  /// else it's encoded to a json String through [jsonEncode] and treated as a String
  ///
  /// [encoding] defaults to [utf8].
  static Request applicationJsonRequest(String method, Uri uri,
      {HeadersMap headers = const {}, Object? body, Encoding? encoding}) {
    final request = Request(method, uri)
      ..headers.addAll(headers)
      ..headers.putIfAbsent('content-type', () => 'application/json');
    if (encoding != null) request.encoding = encoding;
    if (body != null) request.body = body is String ? body : jsonEncode(body);
    return request;
  }

  /// Build an uri from [baseUrl], [url] and query [params]
  ///
  /// Note: if [baseUrl] is not set or [url] does not start with '/', only [url] will be used.
  ///
  /// E.g. buildUrlBase(baseUrl: 'https://my.domain.com', url: '/api/myEndPoint', params: {'q': 'hello'})
  ///      -> 'https://my.domain.com/api/myEndPoint?q=hello'
  static Uri buildUrlBase({String? baseUrl, required String url, QueryParameters params = const {}}) {
    if (baseUrl != null && url.startsWith('/'))
      url = baseUrl.endsWith('/') ? baseUrl + url.substring(1) : baseUrl + url;
    final uri = Uri.parse(url);
    return params.isEmpty ? uri : uri.appendQueryParameters(params);
  }
}

==> 1.1.6-dev.tar.gz: lib/src/x_http_client.dart <==
import 'dart:async';
import 'dart:convert';
import 'dart:typed_data';

import 'package:flutter/foundation.dart';
import 'package:flutterx_http/flutterx_http.dart';
import 'package:flutterx_http/src/transform_request.dart';
import 'package:flutterx_utils/flutterx_utils.dart';
import 'package:http/http.dart';
import 'package:http/src/utils.dart'; // ignore: implementation_imports
import 'package:http_parser/http_parser.dart';

part 'x_http_exception.dart';
part 'x_http_extensions.dart';

typedef OnRequest = FutureOr<BaseRequest> Function(BaseRequest request);
typedef OnResponse = FutureOr<StreamedResponse> Function(StreamedResponse response);
typedef OnError = FutureOr<BaseRequest?> Function(XHTTPException error);
typedef OnMultipartFileProgress = void Function(int fileIndex, int length, int total);

/// Type alias for query parameters map
typedef QueryParameters = Map<String, dynamic>;

/// Type alias for headers map
typedef HeadersMap = Map<String, String>;

class XHttpClient with Loggable {
  final ValueGetter<Client> clientFactory;

  /// Used for building requests url
  final String? baseUrl;

  final OnRequest? onRequest;
  final OnResponse? onResponse;
  final OnError? onError;

  /// Creates a XHttpClient from [client], defaults to Client()
  XHttpClient({
    this.clientFactory = _defaultClientFactory,
    this.baseUrl,
    this.onRequest,
    this.onResponse,
    this.onError,
  });

  Future<Response> head(String url,
          {QueryParameters params = const {}, HeadersMap headers = const {}, CancellationToken? cancel}) =>
      send(applicationJsonRequest('HEAD', buildUrl(url, params: params), headers: headers), cancel: cancel);

  Future<Response> get(String url,
          {QueryParameters params = const {}, HeadersMap headers = const {}, CancellationToken? cancel}) =>
      send(applicationJsonRequest('GET', buildUrl(url, params: params), headers: headers), cancel: cancel);

  Future<Response> post(String url,
          {QueryParameters params = const {},
          HeadersMap headers = const {},
          Object? body,
          Encoding? encoding,
          CancellationToken? cancel}) =>
      send(
          applicationJsonRequest('POST', buildUrl(url, params: params),
              headers: headers, body: body, encoding: encoding),
          cancel: cancel);

  Future<Response> put(String url,
          {QueryParameters params = const {},
          HeadersMap headers = const {},
          Object? body,
          Encoding? encoding,
          CancellationToken? cancel}) =>
      send(
          applicationJsonRequest('PUT', buildUrl(url, params: params),
              headers: headers, body: body, encoding: encoding),
          cancel: cancel);

  Future<Response> patch(String url,
          {QueryParameters params = const {},
          HeadersMap headers = const {},
          Object? body,
          Encoding? encoding,
          CancellationToken? cancel}) =>
      send(
          applicationJsonRequest('PATCH', buildUrl(url, params: params),
              headers: headers, body: body, encoding: encoding),
          cancel: cancel);

  Future<Response> delete(String url,
          {QueryParameters params = const {}, HeadersMap headers = const {}, CancellationToken? cancel}) =>
      send(applicationJsonRequest('DELETE', buildUrl(url, params: params), headers: headers), cancel: cancel);

  Future<Response> multipart(
    String url, {
    String method = 'POST',
    QueryParameters params = const {},
    HeadersMap headers = const {},
    Map<String, String> fields = const {},
    List<MultipartFile> files = const [],
    OnMultipartFileProgress? onProgress,
    CancellationToken? cancel,
  }) {
    BaseRequest request = MultipartRequest(method, buildUrl(url, params: params))
      ..headers.addAll(headers)
      ..headers.putIfAbsent('content-type', () => 'multipart/form-data')
      ..fields.addAll(fields)
      ..files.addAll(files);
    if (onProgress != null) {
      var byteCount = 0;
      var fileIndex = 0;
      request = request.observeProgress((readCount) {
        if (fileIndex >= files.length) return;
        byteCount += readCount;
        final length = files[fileIndex].length;
        if (byteCount > length) {
          onProgress(fileIndex, length, length);
          byteCount -= length;
          fileIndex++;
        } else {
          onProgress(fileIndex, byteCount, length);
        }
      });
    }
    return send(request, cancel: cancel);
  }

  Future<Response> send(BaseRequest request, {CancellationToken? cancel}) async {
    if (onRequest != null) request = await onRequest!(request);
    _printRequest(request);
    final client = clientFactory();
    try {
      var stream = await (cancel == null ? client.send(request) : cancel.attach(client.send(request)));
      if (onResponse != null) stream = await onResponse!(stream);
      final response = await Response.fromStream(stream);
      _printResponse(response);
      if (response.isSuccess) return response;
      final error = XHTTPException(response);
      final retry = onError == null ? null : await onError!(error);
      if (retry != null) return send(retry);
      throw error;
    } finally {
      client.close();
    }
  }

  Uri buildUrl(String url, {QueryParameters params = const {}}) =>
      buildUrlBase(baseUrl: baseUrl, url: url, params: params);

  void _printRequest(BaseRequest request) => log('HTTP Request', () {
        logKV('method', request.method);
        logKV('url', '${request.url.scheme}://${request.url.host}:${request.url.port}${request.url.path}');
        final parameters = request.url.queryParameters;
        if (parameters.isNotEmpty) log('parameters:', () => parameters.forEach(logKV));
        if (request.headers.isNotEmpty) log('headers:', () => request.headers.forEach(logKV));
        if (request.contentLength != null) logKV('contentLength', request.contentLength);
        void _printRequestBody(BaseRequest request) {
          if (request is Request) {
            final body = request.bodyBytes;
            if (body.isNotEmpty) log('body:', () => log(body));
          } else if (request is StreamedRequest) {
            logKV('body', 'stream');
          } else if (request is MultipartRequest) {
            log('body:', () {
              if (request.fields.isNotEmpty) log('fields', () => request.fields.forEach(logKV));
              if (request.files.isNotEmpty) log('files', () => request.files.map((file) => file.filename).forEach(log));
            });
          } else if (request is TransformRequest) {
            logKV('transform', true);
            _printRequestBody(request.source);
          }
        }

        _printRequestBody(request);
      });

  void _printResponse(BaseResponse response) => log('HTTP Response', () {
        final request = response.request;
        if (request != null) {
          logKV('method', request.method);
          logKV('url', request.url);
        }
        logKV('statusCode', response.statusCode);
        logKV('reasonPhrase', response.reasonPhrase);
        if (response.headers.isNotEmpty) log('headers:', () => response.headers.forEach(logKV));
        if (response is Response) {
          final body = response.body;
          if (body.isNotEmpty) log('body:', () => log(body));
        } else if (response is StreamedResponse) {
          log('body: stream');
        }
      });

  /// Creates an HTTP application/json request
  ///
  /// The content-type of the request will default to 'application/json' if not set.
  ///
  /// If provided, [body] sets the body of the request. It can be a [String], any [Object].
  ///
  /// If [body] is a String, it's encoded using [encoding] and used as the body of the request
  /// else it's encoded to a json String through [jsonEncode] and treated as a String
  ///
  /// [encoding] defaults to [utf8].
  static Request applicationJsonRequest(String method, Uri uri,
      {HeadersMap headers = const {}, Object? body, Encoding? encoding}) {
    final request = Request(method, uri)
      ..headers.addAll(headers)
      ..headers.putIfAbsent('content-type', () => 'application/json');
    if (encoding != null) request.encoding = encoding;
    if (body != null) request.body = body is String ? body : jsonEncode(body);
    return request;
  }

  /// Build an uri from [baseUrl], [url] and query [params]
  ///
  /// Note: if [baseUrl] is not set or [url] does not start with '/', only [url] will be used.
  ///
  /// E.g. buildUrlBase(baseUrl: 'https://my.domain.com', url: '/api/myEndPoint', params: {'q': 'hello'})
  ///      -> 'https://my.domain.com/api/myEndPoint?q=hello'
  static Uri buildUrlBase({String? baseUrl, required String url, QueryParameters params = const {}}) {
    if (baseUrl != null && url.startsWith('/'))
      url = baseUrl.endsWith('/') ? baseUrl + url.substring(1) : baseUrl + url;
    final uri = Uri.parse(url);
    return params.isEmpty ? uri : uri.appendQueryParameters(params);
  }

  static Client _defaultClientFactory() => Client();
}
